Program Clique(infile, OUTPUT);
(* version 1.4 *)
const   NmLngth = 10; (* Number of characters in a name *)
        MaxSpp = 50;  (* Max. number of species *)
        MaxChr = 70;  (* Max. number of characters *)
type    Vector = array [1..MaxChr] of boolean;
        NameArray = array [1..MaxSpp, 1..NmLngth] of char;
        DataArray = array [1..MaxSpp] of Vector;
        Matrix = array [1..MaxChr] of Vector;
var     NumSpp, NumChars : integer;
        Name : NameArray;
        Data : DataArray;
        Comp : Matrix; (* The character compatibility matrix *)
        infile : text;
Procedure NewLine (i, j : integer);
          (* Goes to new line if i mod j is zero *)
          begin
          if ((i mod j) = 0)
          then begin
             writeln;
             write('        ');
             end;
          end; (* NewLine *)
Procedure ReadData (var NumSpp, NumChars : integer; var
                    Name : NameArray; var Data : DataArray);
        (* Reads the species names and genetic data *)
        var i, j : integer;
            ch : char;
            badchrs : boolean;
        begin
        badchrs := false;
        readln (infile, NumSpp, NumChars);
        writeln;
        writeln(' ', NumSpp:3,' species,',
                         NumChars:3,' character states');
        writeln;
        writeln(' ', 'Species     Character states');
        writeln(' ', '-------     ----------------');
        for i := 1 to NumSpp do begin
           write(' ');
           for j := 1 to NmLngth do begin
              read (infile, ch); write(ch);
              Name [i, j] := ch;
              end;
           write('  ');
           for j := 1 to NumChars do begin
              if eoln (infile) then readln (infile);
              read (infile, ch); write(ch);
              NewLine( j, 50);
              badchrs := badchrs or ( not (ch in ['0', '1']));
              if (j mod 5 = 0) then write(' ');
              Data [i, j] := (ch = '1');
              end;
           readln (infile); writeln;
           end;
        writeln;
        if badchrs
        then begin
           writeln(' WARNING -- STATES OTHER THAN 0 OR 1 PRESENT IN DATA');
           writeln;
           end;
        end;
Procedure SetUp (var Comp : Matrix);
        (* Sets up the compatibility matrix *)
        var i, j : integer;
        Function Compatible (ch1, ch2 : integer) : boolean;
                (* True if two characters are compatible *)
                var i : integer;
                    info : array [1..4] of boolean;
                begin
                for i := 1 to 4 do info [i] := false;
                for i := 1 to NumSpp do
                   if Data[i,ch1] and Data[i,ch2] then
                      info [1] := true
                   else if Data [i,ch1] and not Data [i,ch2] then
                      info [2] := true
                   else if not Data [i,ch1] and Data [i,ch2] then
                      info [3] := true
                   else info [4] := true;
                Compatible := not (info[1] and info [2] and
                                   info[3] and info [4]);
                end;
        begin (* SetUp *)
        writeln(' ', 'Character Compatibility Matrix');
        writeln(' ', '--------- ------------- ------');
        for i:= 1 to NumChars do begin
           write(' ');
           for j := 1 to i-1 do begin
              if Comp [i, j] then write('1')
              else write('0');
              NewLine( j, 70);
              end;
           Comp [i, i] := true; write('1');
           for j := i+1 to NumChars do begin
              Comp [i, j] := Compatible (i, j);
              if Comp [i, j] then write('1')
              else write('0');
              Comp [j, i] := Comp [i, j];
              end;
           writeln;
           end;
        writeln;
        end;
Procedure DoAll (Chars : Vector; var Processed, Rarer : Vector);
        type    SpVector = array [1..MaxSpp] of integer;
                ChVector = array [1..MaxChr] of integer;
        var     i, j, MaxChars : integer;
                SpOrder : SpVector;
                ChOrder, Count : ChVector;
        Function RarerState (i : integer) : boolean;
                (* True if the rarer state for the i-th
                   character is 1 *)
                var j, ZeroCount, OneCount : integer;
                begin
                ZeroCount :=0; OneCount := 0;
                for j := 1 to  NumSpp do
                   if Data [j, i]
                   then OneCount := OneCount + 1
                   else ZeroCount := ZeroCount + 1;
                RarerState := (OneCount < ZeroCount);
                end;
        Procedure SpSort (var SpOrder : SpVector; i : integer);
                (* Sorts the species according to the rarer
                   state of the i-th character *)
                var j, ptr1, ptr2 : integer;
                    V1, V2 : SpVector;
                begin
                ptr1 := 0; ptr2 := 0;
                for j := 1 to NumSpp do
                   if Data [SpOrder [j], i] then begin
                      ptr1 := ptr1 + 1;
                      V1 [ptr1] := SpOrder [j];
                      end
                   else begin
                      ptr2 := ptr2 + 1;
                      V2 [ptr2] := SpOrder [j];
                      end;
                for j := 1 to ptr1 do SpOrder [j] := V1 [j];
                for j := 1 to ptr2 do SpOrder [j+ptr1] := V2 [j];
                end;
        Procedure Init (var ChOrder, Count : ChVector; var
                        MaxChars : integer);
                (* Initialize vectors and character count *)
                var i, j, temp : integer;
                begin
                MaxChars := 0;
                for i := 1 to NumChars do
                   if Chars [i] then begin
                      MaxChars := MaxChars + 1;
                      ChOrder [MaxChars] := i;
                      temp := 0;
                      for j := 1 to NumSpp do
                         if Data [j, i] then temp := temp + 1;
                      if temp > NumSpp - temp then temp := NumSpp - temp;
                      Count [i] := temp;
                      end;
                end;
        Procedure ChSort (var ChOrder : ChVector; Count : ChVector);
                (* Sorts the characters by cardinality *)
                var j, temp : integer;
                   ordered : boolean;
                begin
                   while not ordered do begin
                      ordered := true;
                      for j := 1 to MaxChars-1 do
                      if Count[ChOrder[j]]<Count[ChOrder[j+1]]
                      then begin
                         ordered := false;
                         temp := ChOrder [j];
                         ChOrder [j] := ChOrder [j+1];
                         ChOrder [j+1] := temp;
                         end;
                     end;
                end;
        Procedure PrintClique (Chars : Vector);
                (* Prints the characters in a clique *)
                var i, j : integer;
                begin
                write(' ', 'Characters: (');
                j := 0;
                for i := 1 to NumChars do begin
                   if Chars [i] then begin
                      write(i:3);
                      j := j + 1;
                      NewLine (j, 25);
                      end;
                   end;
                writeln(')'); writeln;
                end;
        Procedure TreeTable (SpOrder : SpVector; ChOrder : ChVector);
                var i, j, k, temp : integer;
                    ch : char;
                    Spaces : array [0..MaxChr] of integer;
                Function Next (R, C : integer) : integer;
                        begin
                        if C = MaxChars then
                           Next := MaxChars + 1
                        else begin
                           repeat
                            if C < MaxChars then
                               C := C + 1
                           until (C = MaxChars) or
                                (Data [SpOrder[R],ChOrder[C]]);
                           if C = MaxChars then
                            if not Data[SpOrder[R],ChOrder[C]]
                             then C := C + 1;
                           Next := C;
                           end;
                        end;
                begin (* TreeTable *)
                write('  ');
                for i := 1 to MaxChars do begin
                   write(ChOrder [i]:3);
                   end;
                writeln; writeln;
                for i := 0 to MaxChars do
                   Spaces [i] := i;
                for i := 1 to NumSpp do begin
                   write(' ', '!');
                   j := 0;
                   repeat
                    temp := Next (i, j);
                    if (temp <> Spaces [j]) or
                       (temp = MaxChars + 1) then
                       ch := '-'
                    else ch := ' ';
                    for k := j+1 to temp-1 do begin
                       write(ch, ch, ch);
                       NewLine (k, 25);
                         end;
                    if temp <> MaxChars+1 then begin
                       write(ch, ch);
                       if Rarer [ChOrder [temp]]
                       then write('1')
                       else write('0');
                       NewLine (i, 25);
                       end;
                    Spaces [j] := temp;
                    j := temp;
                    NewLine (j, 25);
                   until j = MaxChars + 1;
                   write(' ');
                   for j := 1 to NmLngth do
                      write(Name [SpOrder [i], j]);
                   writeln;
                   end;
                writeln;
                end;
        begin (* DoAll *)
        Init (ChOrder, Count, MaxChars);
        ChSort (ChOrder, Count);
        for i := 1 to NumSpp do SpOrder [i] := i;
         for i := 1 to NumChars do
           if Chars [i] then begin
              if not Processed [i] then begin
                 Rarer [i] := RarerState (i);
                 for j := 1 to NumSpp do
                    Data [j, i] := (Data [j, i] = Rarer [i]);
                 Processed [i] := true;
                 end;
              end;
        for i := MaxChars downto 1 do
           SpSort (SpOrder, ChOrder[i]);
        PrintClique (Chars);
        TreeTable (SpOrder, ChOrder);
        end;
Procedure GetMaxCliques (var Comp : Matrix);
        (* Recursively generates the maximal cliques *)
        var i, TempCount, CharCount, MaxChars : integer;
            Chars, Skip, Temp, Processed, Rarer : Vector;
        Procedure Intersect (V1, V2 : Vector; var V3 : Vector);
                (* Takes the logical intersection V1 and V2 *)
                var i : integer;
                begin
                for i := 1 to NumChars do
                   V3 [i] := V1 [i] and V2 [i];
                end;
        Function Count (V : Vector) : integer;
                (* Counts the 1's in V *)
                var i, TempCount : integer;
                begin
                TempCount := 0;
                for i := 1 to NumChars do
                   if V [i] then TempCount := TempCount + 1;
                Count := TempCount;
                end;
        Procedure Gen1 (i, RowCount : integer; Chars, Skip : Vector);
                (* Determines MaxChar, size of largest clique *)
                var j : integer;
                begin
                if i > 0 then begin
                   Intersect (Chars, Comp [i], Chars);
                   CharCount := Count (Chars);
                   RowCount := RowCount + 1;
                   end
                else CharCount := NumChars;
                for j := i+1 to NumChars do
                   if Chars [j] then
                    if not Skip [j] then
                     if RowCount+NumChars+1-j >= MaxChars then begin
                      Intersect (Chars, Comp [j], Temp);
                      TempCount := Count (Temp);
                      if TempCount = CharCount then begin
                         RowCount := RowCount + 1;
                         Skip [j] := true;
                         end;
                      end;
                If CharCount > RowCount then
                   for j := i+1 to NumChars do begin
                      if Chars [j] then
                       if not Skip [j] then
                        if RowCount+NumChars+1-j >= MaxChars then
                         Gen1 (j, RowCount, Chars, Skip);
                      end
                Else if (CharCount = RowCount)
                      and (CharCount > MaxChars) then
                        MaxChars := CharCount;
                end;
        Procedure Gen2 (i, RowCount : integer; Chars, Skip : Vector);
                (* Finds and outputs the maximal cliques *)
                var j : integer;
                begin
                if i > 0 then begin
                   Intersect (Chars, Comp [i], Chars);
                   CharCount := Count (Chars);
                   RowCount := RowCount + 1;
                   end
                else CharCount := NumChars;
                for j := i+1 to NumChars do
                   if Chars [j] then
                    if not Skip [j] then
                     if RowCount+NumChars+1-j >= MaxChars then begin
                      Intersect (Chars, Comp [j], Temp);
                      TempCount := Count (Temp);
                      if TempCount = CharCount then begin
                         RowCount := RowCount + 1;
                         Skip [j] := true;
                         end;
                      end;
                If CharCount > RowCount then
                   for j := i+1 to NumChars do begin
                      if Chars [j] then
                       if not Skip [j] then
                        if RowCount+NumChars+1-j >= MaxChars then
                         Gen2 (j, RowCount, Chars, Skip);
                      end
                Else if (CharCount = RowCount)
                      and (CharCount = MaxChars) then
                        DoAll (Chars, Processed, Rarer);
                end;
        begin (* GetMaxCliques *)
        writeln(' ', 'Maximal Cliques');
        writeln(' ------- -------');
        MaxChars  := 0;
        for i := 1 to NumChars do begin
           Chars [i] := true;
           Skip [i] := false;
           Processed [i] := false;
           end;
        Gen1 (0, 0, Chars, Skip);
        Gen2 (0, 0, Chars, Skip);
        writeln;
        end;
begin (* Main Program *)
reset (Infile);
ReadData (NumSpp, NumChars, Name, Data);
SetUp (Comp);
GetMaxCliques (Comp);
end.
