PROGRAM CONTML(INFILE,OUTPUT);
  (* VERSION 1.62 *)
CONST MAXCHAR   =  20; (* MAXIMUM NUMBER OF CHARACTERS ALLOWED *)
      MAXSP     =  15; (* MAXIMUM NUMBER OF SPECIES ALLOWED *)
      BONESIZE  =  43; (*THREE PLUS TWICE MAXCHAR*)
      MAXSP2    =  27; (* 2 X MAXSP - 3 *)
      ITERATE   =   2; (* NUMBER OF ITERATIONS TO ADJUST EACH TREE *)
      NAMESIZE  =  10; (* LENGTH IN CHARACTERS OF EACH NAME ON INPUT *)
TYPE  NAYM      =  ARRAY[1..NAMESIZE] OF CHAR;
      VIEW      =  RECORD
                   NAME        :  NAYM;
                   DELTAV      :  REAL;
                   X           :  ARRAY[1..MAXCHAR] OF REAL;
                   END;
      EDGEPTR   =  ^EDGE;
      NODEPTR   =  ^NODE;
      EDGE      =  RECORD
                   FRONT, BACK   :  NODEPTR;
                   ARRANGED      :  BOOLEAN;
                   BACKVIEW,
                   FRONTVIEW     :  VIEW;
                   V             :  REAL;
                   END;
    NODE        =  RECORD
                   INDEX         :  INTEGER;
                   ANCESTOR,
                   FORKLEFT,
                   FORKRIGHT     :  EDGEPTR;
                   END;
    INSERTION   =  RECORD
                   WHERE         :  EDGEPTR;
                   LIKLI         :  REAL;
                   END;
    BONE        =  ARRAY[1..BONESIZE] OF REAL;

VAR INFILE    : TEXT;
    TREE              : EDGEPTR;
    NUMCHAR, NUMSP, I : INTEGER;

PROCEDURE PRINTREE( T : NODEPTR );
   (*FELSENSTEIN, ALGORITHM FOR PRINTING
     CLADOGRAMS ON A LINE PRINTER*)
   CONST C1=2;
         C2=3;
   VAR HORIZ              :  ARRAY[1..100] OF INTEGER;
       ALPHA              :  ARRAY[1..100] OF CHAR;
       SPECIAL, SPECIAL1  :  NAYM;
       I, PAST,
       INDX, OLDIND,
       PLACE1, PLACE2     : INTEGER;

   PROCEDURE MIXORDER (R : NODEPTR);
      VAR I, J : INTEGER;
      BEGIN (*MIXORDER*)
         SPECIAL1:=R^.ANCESTOR^.FRONTVIEW.NAME;
         IF R^.FORKLEFT<>NIL
            THEN MIXORDER(R^.FORKLEFT^.FRONT);
         INDX:=R^.INDEX;
         IF R^.FORKLEFT=NIL
         THEN BEGIN
            IF PAST>0
            THEN BEGIN
               WRITE(OUTPUT,'           ');
               FOR I:=1 TO C2*(PAST-1)+1 DO
                  WRITE(OUTPUT,ALPHA[I]);
               FOR I:=1 TO NAMESIZE DO
                  WRITE(OUTPUT,SPECIAL[I]);
               WRITELN(OUTPUT);
               FOR I:=C2*PAST DOWNTO 2 DO BEGIN 
                  IF ALPHA[I]='-'
                  THEN ALPHA[I]:=' ';
                  IF ALPHA[I] IN ['0'..'9'] 
                  THEN BEGIN
                     IF ALPHA[I-1]=' '
                     THEN ALPHA[I]:='!'
                     ELSE ALPHA[I]:=' ';
                     END;
                  END;
               IF ALPHA[1] IN ['0'..'9'] THEN ALPHA[1]:='!';
               FOR I:=1 TO C1-1 DO BEGIN 
                  WRITE(OUTPUT,'           ');
                  FOR J:=1 TO C2*PAST DO
                      WRITE(OUTPUT,ALPHA[J]);
                  WRITELN(OUTPUT);
                  END;
               END;
            OLDIND:=INDX;
            ALPHA[C2*PAST+1]:='!';
            FOR I:=C2*PAST+2 TO C2*(PAST+1) DO
               ALPHA[I]:=' ';
            PAST:=PAST+1;
            PLACE2:=PAST;
            HORIZ[INDX]:=PAST;
            SPECIAL:=SPECIAL1;
            END;
         IF R^.FORKLEFT<>NIL
         THEN BEGIN
            HORIZ[INDX]:=HORIZ[R^.FORKLEFT^.FRONT^.INDEX];
            MIXORDER(R^.FORKRIGHT^.FRONT);
            END;
         IF R^.ANCESTOR<>NIL
         THEN BEGIN
            IF R^.ANCESTOR=R^.ANCESTOR^.BACK^.FORKRIGHT
            THEN BEGIN
               PLACE1:=HORIZ[R^.ANCESTOR^.BACK^.INDEX];
               FOR I:=C2*(PLACE1-1)+2 TO C2*(PLACE2-1)+1 DO
                  IF NOT (ALPHA[I] IN ['0'..'9']) THEN ALPHA[I]:='-';
               IF TRUNC(R^.ANCESTOR^.BACK^.INDEX/10)<>0
               THEN BEGIN
                  ALPHA[C2*(PLACE1-1)+1]:=CHR(ORD('0')+
                     R^.ANCESTOR^.BACK^.INDEX DIV 10);
                  ALPHA[C2*(PLACE1-1)+2]:=CHR(ORD('0')+
                     R^.ANCESTOR^.BACK^.INDEX MOD 10);
                  END
               ELSE ALPHA[C2*(PLACE1-1)+1]:=CHR(ORD('0')+
                     R^.ANCESTOR^.BACK^.INDEX);
               PLACE2:= PLACE1;
               END;
            END;
      END; (*MIXORDER*)

   BEGIN (*PRINTREE*)
      FOR I:=1 TO NAMESIZE DO
         SPECIAL[I]:=' ';
      PAST:=0;
      WRITELN(OUTPUT);
      MIXORDER(T);
      WHILE T^.FORKLEFT<>NIL DO
         T:=T^.FORKRIGHT^.FRONT;
      SPECIAL:=T^.ANCESTOR^.FRONTVIEW.NAME;
      WRITE(OUTPUT,'           ');
      FOR I:=1 TO C2*(PAST-1)+1 DO WRITE(OUTPUT,ALPHA[I]);
      FOR I:=1 TO NAMESIZE DO
         WRITE(OUTPUT,SPECIAL[I]);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT,'           !');
      WRITE(OUTPUT,'           !');
      FOR I:=1 TO NAMESIZE DO
         WRITE(OUTPUT,TREE^.BACKVIEW.NAME[I]);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT);
   END; (*PRINTREE*)

PROCEDURE MAKETREE (VAR INFILE :  TEXT);
   VAR Q, R, STARTER  :  EDGEPTR;
       PHENO          :  VIEW;
       TEMPLIKLI, 
       A, KL, KU      :  REAL;
       INCOMPLETE,
       BADINSERTION   :  BOOLEAN;
       BESTREE        :  INSERTION;
       BONEINDEX,
       TOTALTREES,
       INDEXER        :  INTEGER;
       CLOSET         :  ARRAY[1..MAXSP2] OF BONE;

   PROCEDURE DISPLAY( STEM : EDGEPTR);
      VAR COUNT : INTEGER;
          BIGV, DV : REAL;
      BEGIN
         BIGV := STEM^.V;
         DV := 0.0;
         IF STEM^.FRONT^.FORKLEFT=NIL
         THEN BEGIN
            WRITE(OUTPUT,' ',STEM^.BACK^.INDEX:3);
            WRITE(OUTPUT,'            ');
            FOR COUNT:=1 TO NAMESIZE DO
               WRITE(OUTPUT,STEM^.FRONTVIEW.NAME[COUNT]);
            FOR COUNT:=1 TO 15-NAMESIZE DO
               WRITE(OUTPUT,' ');
            WRITE(OUTPUT,' ',STEM^.V:7:4);
            DV := STEM^.BACKVIEW.DELTAV;
            END
         ELSE BEGIN 
            IF STEM^.BACK^.ANCESTOR=NIL
            THEN BEGIN
               WRITE(OUTPUT,' ');
               FOR COUNT:=1 TO NAMESIZE DO
                  WRITE(OUTPUT,STEM^.BACKVIEW.NAME[COUNT]);
               FOR COUNT:=1 TO 15-NAMESIZE DO
                  WRITE(OUTPUT,' ');
               WRITE(OUTPUT,STEM^.FRONT^.INDEX:3);
               WRITE(OUTPUT,'            ');
               WRITE(OUTPUT,' ',STEM^.V:7:4);
               DV := STEM^.FRONTVIEW.DELTAV;
               END;
            END;
         IF (STEM^.FRONT^.FORKLEFT<>NIL) AND 
            (STEM^.BACK^.ANCESTOR<>NIL)
         THEN BEGIN
            WRITE(OUTPUT,' ',STEM^.BACK^.INDEX:3);
            WRITE(OUTPUT,'            ');
            WRITE(OUTPUT,STEM^.FRONT^.INDEX:3);
            WRITE(OUTPUT,'            ');
            WRITE(OUTPUT,' ',STEM^.V:7:4);
            DV := STEM^.FRONTVIEW.DELTAV + STEM^.BACKVIEW.DELTAV;
            END;
         BIGV := BIGV + DV;
         WRITELN(OUTPUT,' (',BIGV*KL-DV:10:5,',',BIGV*KU-DV:10:5,' )');
         IF STEM^.FRONT^.FORKLEFT<>NIL THEN
            DISPLAY(STEM^.FRONT^.FORKLEFT);
         IF STEM^.FRONT^.FORKRIGHT<>NIL THEN
            DISPLAY(STEM^.FRONT^.FORKRIGHT);
      END; (*DISPLAY*)

   PROCEDURE UPDATE( TRIANGLE  :  NODEPTR);
      VAR ANCBIGV,
          FLBIGV,
          FRBIGV   :  REAL;
          COUNT    : INTEGER;

      PROCEDURE STEP1;
         VAR COUNT  :  INTEGER;

         FUNCTION VJ(J, K : VIEW) : REAL;
            VAR COUNT : INTEGER;
                SIGMA : REAL;
            BEGIN (*VJ*)
               SIGMA:=0;
               FOR COUNT:=1 TO NUMCHAR DO
                  SIGMA:=SIGMA+SQR(J.X[COUNT]
                           -K.X[COUNT]);
               VJ:=SIGMA/NUMCHAR;
            END; (*VJ*)

         BEGIN(*STEP 1*)
            WITH TRIANGLE^ DO BEGIN 
               ANCBIGV:=0;
               FRBIGV:=0;
               FLBIGV:=0;
               FOR COUNT:=1 TO NUMCHAR DO BEGIN 
                  ANCBIGV:=ANCBIGV+
                            (ANCESTOR^.BACKVIEW.X[COUNT]
                            -FORKLEFT^.FRONTVIEW.X[COUNT])
                            *(ANCESTOR^.BACKVIEW.X[COUNT]
                            -FORKRIGHT^.FRONTVIEW.X[COUNT]);
                  FLBIGV:=FLBIGV+
                            (FORKLEFT^.FRONTVIEW.X[COUNT]
                            -FORKRIGHT^.FRONTVIEW.X[COUNT])
                            *(FORKLEFT^.FRONTVIEW.X[COUNT]
                            -ANCESTOR^.BACKVIEW.X[COUNT]);
                  FRBIGV:=FRBIGV+
                            (FORKRIGHT^.FRONTVIEW.X[COUNT]
                            -FORKLEFT^.FRONTVIEW.X[COUNT])
                            *(FORKRIGHT^.FRONTVIEW.X[COUNT]
                            -ANCESTOR^.BACKVIEW.X[COUNT]);
                  END;
               ANCBIGV:=ANCBIGV/NUMCHAR;
               FLBIGV:=FLBIGV/NUMCHAR;
               FRBIGV:=FRBIGV/NUMCHAR;
               IF ANCBIGV-ANCESTOR^.BACKVIEW.DELTAV<0
               THEN BEGIN
                  ANCBIGV:=ANCESTOR^.BACKVIEW.DELTAV;
                  FLBIGV:=VJ(ANCESTOR^.BACKVIEW,FORKLEFT^.FRONTVIEW)+
                     FORKLEFT^.FRONTVIEW.DELTAV;
                  FRBIGV:=VJ(ANCESTOR^.BACKVIEW,FORKRIGHT^.FRONTVIEW)+
                     FORKRIGHT^.FRONTVIEW.DELTAV;
                  END;
               IF FLBIGV-FORKLEFT^.FRONTVIEW.DELTAV<0
               THEN BEGIN
                  FLBIGV:=FORKLEFT^.FRONTVIEW.DELTAV;
                  ANCBIGV:=VJ(FORKLEFT^.FRONTVIEW,ANCESTOR^.BACKVIEW)+
                     ANCESTOR^.BACKVIEW.DELTAV;
                  FRBIGV:=VJ(FORKLEFT^.FRONTVIEW,FORKRIGHT^.FRONTVIEW)+
                     FORKRIGHT^.FRONTVIEW.DELTAV;
                  END;
               IF FRBIGV-FORKRIGHT^.FRONTVIEW.DELTAV<0
               THEN BEGIN
                  FRBIGV:=FORKRIGHT^.FRONTVIEW.DELTAV;
                  ANCBIGV:=VJ(FORKRIGHT^.FRONTVIEW,ANCESTOR^.BACKVIEW)+
                     ANCESTOR^.BACKVIEW.DELTAV;
                  FLBIGV:=VJ(FORKRIGHT^.FRONTVIEW,FORKLEFT^.FRONTVIEW)+
                     FORKLEFT^.FRONTVIEW.DELTAV;
                  END;
               END;
         END;(*STEP 1*)

      PROCEDURE STEP2;
         BEGIN(*STEP 2*)
            WITH TRIANGLE^ DO BEGIN 
               ANCESTOR^.FRONTVIEW.DELTAV:=FLBIGV*FRBIGV
                                 /(FLBIGV+FRBIGV);
               FORKLEFT^.BACKVIEW.DELTAV:=FRBIGV*ANCBIGV
                                /(FRBIGV+ANCBIGV);
               FORKRIGHT^.BACKVIEW.DELTAV:=FLBIGV*ANCBIGV
                                 /(FLBIGV+ANCBIGV);
               END;
         END; (*STEP 2*)

      PROCEDURE STEP3;
         BEGIN (*STEP 3*)
            WITH TRIANGLE^ DO BEGIN 
               ANCESTOR^.V:=ANCBIGV-ANCESTOR^.BACKVIEW.DELTAV;
               FORKLEFT^.V:=FLBIGV-FORKLEFT^.FRONTVIEW.DELTAV;
               FORKRIGHT^.V:=FRBIGV-FORKRIGHT^.FRONTVIEW.DELTAV;
               END;
         END;(*STEP 3*)

      PROCEDURE STEP4;
         VAR COUNT  :  INTEGER;
         BEGIN(*STEP 4*)
            WITH TRIANGLE^ DO BEGIN 
               FOR COUNT:=1 TO NUMCHAR DO BEGIN 
                  ANCESTOR^.FRONTVIEW.X[COUNT]:=
                      (FLBIGV*FORKRIGHT^.FRONTVIEW.X[COUNT]
                      +FRBIGV*FORKLEFT^.FRONTVIEW.X[COUNT])
                      /(FRBIGV+FLBIGV);
                  FORKLEFT^.BACKVIEW.X[COUNT]:=
                      (ANCBIGV*FORKRIGHT^.FRONTVIEW.X[COUNT]
                      +FRBIGV*ANCESTOR^.BACKVIEW.X[COUNT])
                      /(FRBIGV+ANCBIGV);
                  FORKRIGHT^.BACKVIEW.X[COUNT]:=
                      (ANCBIGV*FORKLEFT^.FRONTVIEW.X[COUNT]
                      +FLBIGV*ANCESTOR^.BACKVIEW.X[COUNT])
                      /(FLBIGV+ANCBIGV);
                  END;
               END;
         END;(*STEP 4*)

      BEGIN 
         STEP1;
         IF (ANCBIGV+FRBIGV=0) OR
            (FRBIGV +FLBIGV=0) OR
            (FLBIGV+ANCBIGV=0)
            THEN BADINSERTION:=TRUE;
         IF NOT BADINSERTION
         THEN BEGIN
            STEP2;
            STEP3;
            STEP4;
            END;
      END;(*UPDATE*)

   PROCEDURE SMOOTH(TRIANGLE : NODEPTR);
      BEGIN(*SMOOTH*)
         UPDATE(TRIANGLE);
         IF TRIANGLE^.FORKRIGHT^.FRONT^.FORKLEFT<>NIL
            THEN SMOOTH(TRIANGLE^.FORKRIGHT^.FRONT);
         IF TRIANGLE^.FORKLEFT^.FRONT^.FORKLEFT<>NIL
            THEN SMOOTH(TRIANGLE^.FORKLEFT^.FRONT);
      END; (*SMOOTH*)

   PROCEDURE STORE(STEM : EDGE);
      VAR COUNT : INTEGER;
      BEGIN(*STORE*)
         BONEINDEX:=BONEINDEX+1;
         WITH STEM DO BEGIN 
            FOR COUNT:=1 TO NUMCHAR DO BEGIN 
               CLOSET[BONEINDEX][2*COUNT-1]:=FRONTVIEW.X[COUNT];
               CLOSET[BONEINDEX][2*COUNT]:=BACKVIEW.X[COUNT];
               END;
            CLOSET[BONEINDEX][2*NUMCHAR+1]:=FRONTVIEW.DELTAV;
            CLOSET[BONEINDEX][2*NUMCHAR+2]:=BACKVIEW.DELTAV;
            CLOSET[BONEINDEX][2*NUMCHAR+3]:=V;
            END;
         IF STEM.FRONT^.FORKLEFT<>NIL
            THEN STORE(STEM.FRONT^.FORKLEFT^);
         IF STEM.FRONT^.FORKRIGHT<>NIL
            THEN STORE(STEM.FRONT^.FORKRIGHT^);
      END;(*STORE*)

   PROCEDURE GETREE(VAR STEM : EDGE);
      VAR COUNT : INTEGER;
      BEGIN(*GETREE*)
         BONEINDEX:=BONEINDEX+1;
         WITH STEM DO BEGIN 
            FOR COUNT:=1 TO NUMCHAR DO BEGIN 
               FRONTVIEW.X[COUNT]:=CLOSET[BONEINDEX][2*COUNT-1];
               BACKVIEW.X[COUNT]:=CLOSET[BONEINDEX][2*COUNT];
               END;
            FRONTVIEW.DELTAV:=CLOSET[BONEINDEX][2*NUMCHAR+1];
            BACKVIEW.DELTAV:=CLOSET[BONEINDEX][2*NUMCHAR+2];
            V:=CLOSET[BONEINDEX][2*NUMCHAR+3];
            END;
         IF STEM.FRONT^.FORKLEFT<>NIL
            THEN GETREE(STEM.FRONT^.FORKLEFT^);
         IF STEM.FRONT^.FORKRIGHT<>NIL
            THEN GETREE(STEM.FRONT^.FORKRIGHT^);
      END;(*GETREE*)

   FUNCTION EVALUATE  :  REAL;
      VAR COUNT  :  INTEGER;
          NATLOG, SIGMA,
          TEMPLIKLI,
          CAPV   :  REAL;

      FUNCTION LGNATURAL ( X : REAL ) : REAL;
         VAR Y : REAL;
         BEGIN
           Y := LN(X);
           IF X<0 THEN Y:=-ABS(Y);
           LGNATURAL := Y;
         END; (*LGNATURAL*)

      PROCEDURE SUMLIKLI( T : NODEPTR );
         VAR COUNT : INTEGER;
         BEGIN
            IF T^.FORKLEFT^.FRONT^.FORKLEFT<>NIL
               THEN SUMLIKLI ( T^.FORKLEFT^.FRONT);
            IF T^.FORKRIGHT^.FRONT^.FORKLEFT<>NIL
               THEN SUMLIKLI ( T^.FORKRIGHT^.FRONT);
            SIGMA:=0;
            FOR COUNT:=1 TO NUMCHAR DO
               SIGMA:=SIGMA+SQR(T^.FORKLEFT^.FRONTVIEW.X[COUNT]
                        -T^.FORKRIGHT^.FRONTVIEW.X[COUNT]);
            CAPV:=T^.FORKLEFT^.V+
                T^.FORKRIGHT^.V+
                T^.FORKLEFT^.FRONTVIEW.DELTAV+
                T^.FORKRIGHT^.FRONTVIEW.DELTAV;
            SIGMA:=SIGMA/CAPV/2;
            NATLOG:=LGNATURAL(CAPV);
            TEMPLIKLI:=TEMPLIKLI-SIGMA-NUMCHAR/2*NATLOG;
         END; (*SUMLIKLI*)

      BEGIN (*EVALUATE*)
         TEMPLIKLI:=0;
         SUMLIKLI(TREE^.FRONT);
         SIGMA:=0;
         FOR COUNT:=1 TO NUMCHAR DO
            SIGMA:=SIGMA+SQR(TREE^.FRONTVIEW.X[COUNT]
                      -TREE^.BACKVIEW.X[COUNT]);
         CAPV:=TREE^.FRONTVIEW.DELTAV+TREE^.V;
         SIGMA:=SIGMA/CAPV/2;
         NATLOG:=LGNATURAL(CAPV);
         TEMPLIKLI:=TEMPLIKLI-SIGMA-NUMCHAR/2*NATLOG;
         EVALUATE:=TEMPLIKLI;
      END; (*EVALUATE*)

   PROCEDURE MAKEVIEW;
      VAR COUNT : INTEGER;
      BEGIN (* MAKEVIEW *)
         WRITE(OUTPUT,' ');
         FOR COUNT:=1 TO NAMESIZE DO BEGIN 
            READ(INFILE,PHENO.NAME[COUNT]);
            WRITE(OUTPUT,PHENO.NAME[COUNT]);
            END;
         WRITE(OUTPUT,'   ');
         FOR COUNT:=1 TO NUMCHAR DO BEGIN 
            IF EOLN(INFILE) THEN READLN(INFILE);
            READ(INFILE,PHENO.X[COUNT]);
            IF (COUNT <> 1) AND ((COUNT MOD 6) = 1)
            THEN BEGIN
               WRITELN(OUTPUT);
               WRITE(OUTPUT,'                ');
               END;
            WRITE(OUTPUT,PHENO.X[COUNT]:10:5);
            END;
         READLN(INFILE);
         WRITELN(OUTPUT);
      END; (* MAKEVIEW *)

   PROCEDURE BUILDQR;
      VAR SGMNT    : EDGEPTR;
          TRIANGLE : NODEPTR;
          COUNT    : INTEGER;
      BEGIN (*BUILDQR*)
         NEW(SGMNT);
         SGMNT^.ARRANGED := FALSE;
         NEW(TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         FOR COUNT := 1 TO NAMESIZE DO
            SGMNT^.BACKVIEW.NAME[COUNT]:=' ';
         MAKEVIEW;
         PHENO.DELTAV:=0;
         SGMNT^.FRONTVIEW:=PHENO;
         SGMNT^.FRONT:=TRIANGLE;
         TRIANGLE^.ANCESTOR:=SGMNT;
         TRIANGLE^.FORKLEFT:=NIL;
         TRIANGLE^.FORKRIGHT:=NIL;
         NEW(TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         TRIANGLE^.FORKLEFT:=SGMNT;
         SGMNT^.BACK:=TRIANGLE;
         NEW(SGMNT);
         SGMNT^.ARRANGED := FALSE;
         FOR COUNT:=1 TO NAMESIZE DO BEGIN 
            SGMNT^.FRONTVIEW.NAME[COUNT]:=' ';
            SGMNT^.BACKVIEW.NAME[COUNT]:=' ';
            END;
         TRIANGLE^.FORKRIGHT:=SGMNT;
         TRIANGLE^.ANCESTOR:=NIL;
         SGMNT^.FRONT:=NIL;
         SGMNT^.BACK:=TRIANGLE;
         Q:=TRIANGLE^.FORKLEFT;
         R:=TRIANGLE^.FORKRIGHT;
      END; (*BUILDQR*)

   PROCEDURE BUILDSIMPLETREE;
      VAR SGMNT    :   EDGEPTR;
          TRIANGLE :   NODEPTR;
          COUNT    :   INTEGER;
      BEGIN (*BUILDSIMPLETREE*)
         INDEXER:=0;
         BUILDQR;
         NEW (SGMNT);
         SGMNT^.ARRANGED := FALSE;
         NEW (TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         TRIANGLE^.ANCESTOR:=NIL;
         TRIANGLE^.FORKRIGHT:=NIL;
         SGMNT^.BACK:=TRIANGLE;
         SGMNT^.FRONT:=Q^.BACK;
         FOR COUNT:=1 TO NAMESIZE DO
            SGMNT^.FRONTVIEW.NAME[COUNT]:=' ';
         Q^.BACK^.ANCESTOR:=SGMNT;
         MAKEVIEW;
         PHENO.DELTAV:=0;
         SGMNT^.BACKVIEW:=PHENO;
         TRIANGLE^.FORKLEFT:=SGMNT;
         NEW(TRIANGLE);
         INDEXER:=INDEXER+1;
         TRIANGLE^.INDEX:=INDEXER;
         R^.FRONT:=TRIANGLE;
         TRIANGLE^.ANCESTOR:=R;
         MAKEVIEW;
         PHENO.DELTAV:=0;
         R^.FRONTVIEW:=PHENO;
         TRIANGLE^.FORKLEFT:=NIL;
         TRIANGLE^.FORKRIGHT:=NIL;
         BADINSERTION:=FALSE;
         UPDATE(R^.BACK);
         TREE:=SGMNT;
                 (*SET TREE TO POINT TO "QUASI-ANCESTRAL
                  SGMNT:  RECURSION ROUTINES
                  NEED SOMEWHERE TO BEGIN AND
                  THIS SOMEWHERE IS UNIQUE BECAUSE
                  OF THE POLARITY OF THE SEGMENTS*)
      END; (*BUILDSIMPLETREE*)

   PROCEDURE INSERT (P, R  : EDGEPTR);
      VAR COUNT : INTEGER;
      BEGIN (*INSERT*)
         R^.FRONT:=P^.FRONT;
         P^.FRONT:=R^.BACK;
         R^.BACK^.ANCESTOR:=P;
         R^.FRONT^.ANCESTOR:=R;
         R^.FRONTVIEW:=P^.FRONTVIEW;
         R^.ARRANGED:=P^.ARRANGED;
         IF R^.FRONT^.FORKLEFT<>NIL
         THEN BEGIN
            UPDATE(R^.BACK);
            IF NOT BADINSERTION
               THEN UPDATE(R^.FRONT);
            END
         ELSE UPDATE(R^.BACK);
      END; (*INSERT*)

   PROCEDURE DELETE(R : EDGEPTR);
      VAR P : EDGEPTR;
      BEGIN (*DELETE*)
         P:=R^.BACK^.ANCESTOR;
         P^.FRONT:= R^.FRONT;
         R^.FRONT:=NIL;
         P^.FRONT^.ANCESTOR:=P;
         R^.BACK^.ANCESTOR:=NIL;
         P^.FRONTVIEW:=R^.FRONTVIEW;
         IF P^.FRONT^.FORKLEFT<>NIL
            THEN UPDATE(P^.FRONT);
      END; (*DELETE*)

   FUNCTION LOOKFORNEG : BOOLEAN;
      VAR NEG : BOOLEAN;

      PROCEDURE SCAN(T : NODEPTR);
         BEGIN (*SCAN*)
            IF (T^.FORKLEFT^.FRONTVIEW.DELTAV+
                T^.FORKRIGHT^.FRONTVIEW.DELTAV+
                T^.FORKLEFT^.V+T^.FORKRIGHT^.V<=0)
            THEN NEG:=TRUE;
            IF T^.FORKLEFT^.FRONT^.FORKLEFT<>NIL
               THEN SCAN(T^.FORKLEFT^.FRONT);
            IF T^.FORKRIGHT^.FRONT^.FORKLEFT<>NIL
               THEN SCAN(T^.FORKRIGHT^.FRONT);
         END; (*SCAN*)

      BEGIN(*LOOKFORNEG*)
         NEG:=FALSE;
         IF TREE^.FRONTVIEW.DELTAV+TREE^.V<=0
            THEN NEG:=TRUE;
         IF NOT NEG THEN SCAN(TREE^.FRONT);
         LOOKFORNEG:=NEG;
      END; (*LOOKFORNEG*)

   PROCEDURE PLACENEWTIP;
      VAR COUNT : INTEGER;

      PROCEDURE SEARCH (STEM : EDGEPTR);
         VAR COUNT  :  INTEGER;
         BEGIN (*SEARCH*)
            IF STEM^.FRONT^.FORKLEFT<>NIL
            THEN BEGIN
               TOTALTREES:=TOTALTREES+1;
               BONEINDEX:=0;
               GETREE(TREE^);
               BADINSERTION:=FALSE;
               INSERT (STEM^.FRONT^.FORKLEFT, R);
               IF NOT BADINSERTION
                  THEN FOR COUNT:=1 TO ITERATE DO SMOOTH(TREE^.FRONT);
               IF (NOT LOOKFORNEG) AND (NOT BADINSERTION)
               THEN BEGIN
                  TEMPLIKLI:=EVALUATE;
                  DELETE(R);
                  IF TEMPLIKLI>BESTREE.LIKLI
                  THEN BEGIN
                     BESTREE.WHERE:=STEM^.FRONT^.FORKLEFT;
                     BESTREE.LIKLI:=TEMPLIKLI;
                     END;
                  END
               ELSE DELETE(R);
               SEARCH (STEM^.FRONT^.FORKLEFT);
               END;
            IF STEM^.FRONT^.FORKRIGHT<>NIL
            THEN BEGIN
               TOTALTREES:=TOTALTREES+1;
               BONEINDEX:=0;
               GETREE(TREE^);
               BADINSERTION:=FALSE;
               INSERT (STEM^.FRONT^.FORKRIGHT, R);
               IF NOT BADINSERTION
                  THEN FOR COUNT:=1 TO ITERATE DO SMOOTH(TREE^.FRONT);
               IF (NOT LOOKFORNEG) AND (NOT BADINSERTION)
               THEN BEGIN
                  TEMPLIKLI:=EVALUATE;
                  DELETE(R);
                  IF TEMPLIKLI>BESTREE.LIKLI
                  THEN BEGIN
                     BESTREE.WHERE:=STEM^.FRONT^.FORKRIGHT;
                     BESTREE.LIKLI:=TEMPLIKLI;
                     END;
                  END
               ELSE DELETE(R);
               SEARCH (STEM^.FRONT^.FORKRIGHT);
               END;
         END; (*SEARCH*)

      BEGIN (*PLACENEWTIP*)
         BONEINDEX:=0;
         GETREE(TREE^);
         BADINSERTION:=FALSE;
         INSERT(TREE, R);
         TOTALTREES:=TOTALTREES+1;
         IF NOT BADINSERTION
            THEN FOR COUNT:=1 TO ITERATE DO SMOOTH(TREE^.FRONT);
         IF (NOT LOOKFORNEG) AND (NOT BADINSERTION)
         THEN BEGIN
            TEMPLIKLI:=EVALUATE;
            DELETE(R);
            BESTREE.WHERE:=TREE;
            BESTREE.LIKLI:=TEMPLIKLI;
            END
         ELSE BEGIN 
            DELETE(R);
            BESTREE.WHERE:=TREE;
            BESTREE.LIKLI:=-5000.0;
            END;
         SEARCH(TREE);
         IF BESTREE.LIKLI=-5000.0
            THEN WRITELN(OUTPUT,' ERROR IN PLACING TIP');
         INSERT(BESTREE.WHERE, R);
         FOR COUNT:=1 TO ITERATE DO SMOOTH(TREE^.FRONT);
         BONEINDEX:=0;
         STORE(TREE^);
      END;(*PLACENEWTIP*)

   PROCEDURE INITIALIZE (STEM : EDGEPTR);
      BEGIN (*INITIALIZE*)
         STEM^.ARRANGED:=FALSE;
         IF STEM^.FRONT^.FORKLEFT<>NIL
            THEN INITIALIZE (STEM^.FRONT^.FORKLEFT);
         IF STEM^.FRONT^.FORKRIGHT<>NIL
            THEN INITIALIZE (STEM^.FRONT^.FORKRIGHT);
      END; (*INITIALIZE*)

   PROCEDURE REARRANGE(STEM  :  EDGEPTR);
      VAR RERANGEABLE   : BOOLEAN;
          REAR      : EDGEPTR;
          POSITION,
          RO, COUNT   : INTEGER;

      PROCEDURE ROTATE(VAR STEM : EDGEPTR; FULL : BOOLEAN);
         VAR  RIGHT    : BOOLEAN;
              COUNT    : INTEGER;
              TEMP,
              ROOT   : EDGEPTR;
         BEGIN
            ROOT:=STEM^.BACK^.ANCESTOR;
            RIGHT:=FALSE;
            IF STEM^.BACK^.FORKRIGHT=STEM
               THEN RIGHT:=TRUE;
            IF NOT RIGHT 
            THEN BEGIN
               TEMP:=ROOT^.FRONT^.FORKLEFT;
               ROOT^.FRONT^.FORKLEFT:=ROOT^.FRONT^.FORKRIGHT;
               ROOT^.FRONT^.FORKRIGHT:=TEMP;
               END;
            DELETE(STEM);
            IF FULL
               THEN UPDATE(ROOT^.FRONT);
            INSERT(ROOT^.FRONT^.FORKLEFT,STEM);
            STEM:=ROOT^.FRONT^.FORKLEFT;
            IF RIGHT
            THEN BEGIN
               TEMP:=ROOT^.FRONT^.FORKLEFT;
               ROOT^.FRONT^.FORKLEFT:=ROOT^.FRONT^.FORKRIGHT;
               ROOT^.FRONT^.FORKRIGHT:=TEMP;
               END;
         END; (*ROTATE*)

      BEGIN (*REARRANGE*)
         RERANGEABLE:=FALSE;
         IF (STEM^.FRONT^.FORKLEFT<>NIL) THEN RERANGEABLE:=TRUE;
         IF RERANGEABLE
         THEN BEGIN
            POSITION:=1;
            FOR RO:=1 TO 2 DO BEGIN 
               BADINSERTION:=FALSE;
               TOTALTREES:=TOTALTREES+1;
               ROTATE(STEM,TRUE);
               IF NOT BADINSERTION
                  THEN FOR COUNT:=1 TO ITERATE DO SMOOTH(TREE^.FRONT);
               IF (NOT LOOKFORNEG) AND (NOT BADINSERTION)
               THEN BEGIN
                  TEMPLIKLI:=EVALUATE;
                  IF TEMPLIKLI>BESTREE.LIKLI
                  THEN BEGIN
                     BONEINDEX:=0;
                     STORE(TREE^);
                     POSITION:=RO+1;
                     BESTREE.LIKLI:=TEMPLIKLI;
                     END;
                  END;
               END;
            FOR RO:=1 TO POSITION DO ROTATE(STEM,FALSE);
            IF POSITION<>1 THEN INCOMPLETE:=TRUE;
            BONEINDEX:=0;
            GETREE(TREE^);
            STEM^.ARRANGED:=TRUE;
            IF STEM^.BACK^.FORKRIGHT=STEM
            THEN REAR:=STEM^.BACK^.FORKLEFT
            ELSE REAR:=STEM^.BACK^.FORKRIGHT;
            IF (NOT STEM^.FRONT^.FORKLEFT^.ARRANGED)
               THEN REARRANGE (STEM^.FRONT^.FORKLEFT);
            IF (NOT STEM^.FRONT^.FORKRIGHT^.ARRANGED)
               THEN REARRANGE (STEM^.FRONT^.FORKRIGHT);
            IF (NOT REAR^.ARRANGED)
               THEN REARRANGE (REAR);
            END
         ELSE STEM^.ARRANGED:=TRUE;
      END;  (*REARRANGE*)

   BEGIN (*MAKETREE*)
      READLN(INFILE, NUMSP, NUMCHAR);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT,' SPECIES                PHENOTYPES');
      WRITELN(OUTPUT,' -------                ----------');
      BUILDSIMPLETREE;
      TOTALTREES:=1;
      A := 2/(9*NUMCHAR);
      KL := 1 - A - 1.95996*SQRT(A);
      KU := 1 - A + 1.95996*SQRT(A);
      KL := KL*KL*KL;
      KU := KU*KU*KU;
      IF NUMCHAR = 1 THEN KL := 0.0000982;
      IF NUMCHAR = 2 THEN KL := 0.0253;
      BONEINDEX:=0;
      STORE(TREE^);
      BUILDQR;
      PLACENEWTIP;
      PRINTREE (TREE^.FRONT);
      WRITELN(OUTPUT,' FROM             TO              LENGTH');
      WRITELN(OUTPUT,' ----             --              ------');
      DISPLAY(TREE);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT,'   TOTAL TREES INSPECTED = ',TOTALTREES);
      WRITELN(OUTPUT);
      WRITELN(OUTPUT,'   LOG LIKELIHOOD = ',BESTREE.LIKLI:9:4);
      I := 4;
      WHILE I<NUMSP DO BEGIN 
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT,' READ IN NEXT SPECIES:');
         BUILDQR;
         I := I + 1;
         PLACENEWTIP;
         INCOMPLETE:=TRUE;
         WHILE INCOMPLETE DO BEGIN 
            INCOMPLETE:=FALSE;
            INITIALIZE (TREE);
            STARTER:=TREE^.FRONT^.FORKLEFT;
            IF STARTER^.FRONT^.FORKLEFT=NIL
               THEN STARTER:=TREE^.FRONT^.FORKRIGHT;
            REARRANGE (STARTER);
            END;
         PRINTREE (TREE^.FRONT);
         WRITELN(OUTPUT,' FROM           TO               LENGTH');
         WRITELN(OUTPUT,' ----           --               ------');
         DISPLAY(TREE);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT,'   TOTAL TREES INSPECTED = ',TOTALTREES);
         WRITELN(OUTPUT);
         WRITELN(OUTPUT,'   LOG LIKELIHOOD = ',BESTREE.LIKLI:9:4);
         END;
      BONEINDEX:=0;
      GETREE(TREE^);
   END; (*MAKETREE*)

BEGIN (*MAIN PROGRAM*)
   RESET(INFILE);
   MAKETREE (INFILE);
END. (*MAIN PROGRAM*)
