PROGRAM DNACOMPAT(INFILE, OUTPUT);
 (* VERSION 1.4 *)
CONST MAXSZ = 50;  (* SIZE OF POINTER ARRAY.  >= 2*MAXSP - 1  *)
  MAXSP = 25;       (* MAXIMUM NUMBER OF SPECIES               *)
  NMLNGTH = 10;     (* NUMBER OF CHARACTERS IN SPECIES NAME    *)
  MAXSITES = 500;   (* MAX. NUMBER OF DNA BASE SITES PER SEQUENCE *)
TYPE POINTER = ^NODE;           (* NODES WILL FORM BINARY TREE *)
  POINTARRAY = ARRAY [1..MAXSZ] OF POINTER;
  BASES = (A, C, G, U);
  BASET = SET OF BASES;
  NODE = RECORD      (* DESCRIBES A TIP SPECIES OR AN ANCESTOR *)
     ANCESTOR, LEFTDESC, RTDESC : POINTER;
     INDEX : INTEGER;
     TIP : BOOLEAN;        (* PRESENT SPECIES ARE TIPS OF TREE *)
     STATESET : ARRAY [1..MAXSITES] OF BASET;  (* CHARACTERISTICS *)
     END;
VAR ROOT : POINTER;
  INFILE : TEXT;
  I, J, K, L, SPP, SITES, NONODES : INTEGER;
    (* SPP = NUMBER OF SPECIES; SITES = NUMBER OF SITES NEEDED
       NONODES = NUMBER OF NODES IN TREE *)
  LIKE : INTEGER;                    (* "LIKELIHOOD" OF A TREE *)
  TREENODE : POINTARRAY;      (* POINTERS TO ALL NODES IN TREE *)
  NAME : ARRAY[1..MAXSP, 1..NMLNGTH] OF CHAR;(*NAMES OF SPECIES*)
  NS : BASET; (* TEMPORARY BASE SET FOR INPUT *)
  CHARSTATE : CHAR;      (* TEMPORARY FOR BASE ON INPUT *)
  NECSTEPS, STEPS : ARRAY[1..MAXSITES] OF INTEGER;
PROCEDURE ADD (VAR BELOW, NEWTIP, NEWFORK : POINTER);
  (* INSERTS THE NODES NEWFORK AND ITS LEFT DESCENDANT, NEWTIP,
     TO THE TREE.  BELOW BECOMES NEWFORK'S RIGHT DESCENDANT *)
  BEGIN (* ADD *)
     NEWFORK^.ANCESTOR := BELOW^.ANCESTOR;
     BELOW^.ANCESTOR := NEWFORK;
     NEWTIP^.ANCESTOR := NEWFORK;
     NEWFORK^.RTDESC := BELOW;
     NEWFORK^.LEFTDESC := NEWTIP;
     IF NEWFORK^.ANCESTOR <> NIL THEN BEGIN
        IF NEWFORK^.ANCESTOR^.LEFTDESC = BELOW
        THEN NEWFORK^.ANCESTOR^.LEFTDESC := NEWFORK
        ELSE NEWFORK^.ANCESTOR^.RTDESC := NEWFORK;
        END
  END; (* ADD *)
PROCEDURE REMOVE (VAR ITEM, FORK : POINTER);
  (* REMOVES NODES ITEM AND ITS ANCESTOR, FORK, FROM THE TREE.
     THE NEW DESCENDANT OF FORK'S ANCESTOR IS MADE TO BE
     FORK'S SECOND DESCENDANT (OTHER THAN ITEM).
     ALSO RETURNS POINTERS TO THE DELETED NODES, ITEM AND FORK *)
  VAR OTHER, BELOW : POINTER;
  BEGIN (* REMOVE *)
     IF ITEM^.ANCESTOR = NIL
     THEN FORK := NIL
     ELSE BEGIN
        FORK := ITEM^.ANCESTOR;
        IF ITEM = FORK^.LEFTDESC
        THEN OTHER := FORK^.RTDESC
        ELSE OTHER := FORK^.LEFTDESC;
        BELOW := FORK^.ANCESTOR;
        OTHER^.ANCESTOR := BELOW;
        IF BELOW <> NIL THEN BEGIN
           IF BELOW^.LEFTDESC = FORK
           THEN BELOW^.LEFTDESC := OTHER
           ELSE BELOW^.RTDESC := OTHER;
           FORK^.ANCESTOR := NIL;
           END;
        FORK^.LEFTDESC := NIL;
        FORK^.RTDESC := NIL;
        ITEM^.ANCESTOR := NIL;
        END;
  END; (* REMOVE *)
PROCEDURE PRINTREE (P: POINTER);
  (* PRINTS A BINARY TREE *)
  CONST C1=2; C2=3;
  VAR HORIZ: ARRAY[1..MAXSZ] OF INTEGER;
     ALPHA: ARRAY[1..100] OF CHAR;
     I, J, K, PAST, INDX, OLDIND, PLACE1, PLACE2, C22 : INTEGER;
  PROCEDURE MIXORDER (R: POINTER);
     (* THIS IS A MIXED-ORDER TREE TRAVERSAL WHICH DOES MOST OF
        THE WORK OF PRINTREE *)
     VAR I,J : INTEGER;
     BEGIN (* MIXORDER *)
        IF NOT R^.TIP THEN
           MIXORDER (R^.LEFTDESC);
        INDX := R^.INDEX;
        IF R^.TIP THEN BEGIN
           IF PAST > 0 THEN BEGIN
              WRITE(' ');
              FOR I:=1 TO C22*(PAST-1)+1 DO WRITE(ALPHA[I]);
              FOR I:=1 TO NMLNGTH DO WRITE(NAME[OLDIND,I]);
              WRITELN;
              FOR I:= C22*PAST DOWNTO 2 DO BEGIN
                 IF ALPHA[I] = '-' THEN ALPHA[I] := ' ';
                 IF ALPHA[I] = '*' THEN BEGIN
                    IF ALPHA[I-1] = ' '
                       THEN ALPHA[I] := '!'
                       ELSE ALPHA[I] := ' ';
                    END;
                 END;
              IF ALPHA[1] = '*' THEN ALPHA[1] := '!';
              FOR I:=1 TO C1-1 DO BEGIN
                 WRITE(' ');
                 FOR J:=1 TO C22*PAST DO WRITE(ALPHA[J]);
                 WRITELN;
                 END
              END;
           OLDIND := INDX;
           ALPHA[C22*PAST + 1] := '!';
           FOR I:= C22*PAST+2 TO C22*(PAST+1) DO ALPHA[I] := ' ';
           PAST := PAST +1;
           PLACE2 := PAST;
           HORIZ[INDX] := PAST;
           END;
        IF NOT R^.TIP THEN BEGIN
           HORIZ[INDX] := HORIZ[R^.LEFTDESC^.INDEX];
           MIXORDER (R^.RTDESC);
           END;
        IF R^.ANCESTOR <> NIL THEN BEGIN
           IF R = R^.ANCESTOR^.RTDESC THEN BEGIN
              PLACE1 := HORIZ[R^.ANCESTOR^.INDEX];
              FOR I:= C22*(PLACE1-1)+2 TO C22*(PLACE2-1)+1 DO
                 IF ALPHA[I] <> '*' THEN ALPHA[I] := '-';
              ALPHA[C22*(PLACE1-1)+1] := '*';
              PLACE2 := PLACE1;
              END;
        END;
     END; (* MIXORDER *)
  BEGIN (* PRINTREE *)
     C22 := 100 DIV SPP;
     IF C22 > C2 THEN C22 := C2;
     PAST := 0;
     WRITELN;
     WRITELN;
     MIXORDER (P);
     WRITE(' ');
     ALPHA[1] := 'L';
     FOR I:=1 TO C22*(PAST-1)+1 DO WRITE(ALPHA[I]);
     FOR I:= 1 TO NMLNGTH DO WRITE(NAME[INDX,I]);
     WRITELN;
     WRITELN;
     WRITELN(' REMEMBER, THIS TREE IS UNROOTED');
     WRITELN;
     WRITELN(' COMPATIBILITY (Y OR N) OF EACH SITE WITH THIS TREE:');
     WRITELN;
     WRITE(' SITE:');
     FOR I := 0 TO 9 DO WRITE(I:1);
     WRITELN;
     WRITE('     *');
     WRITE('----------');
     WRITELN;
     FOR I := 1 TO (SITES DIV 10 + 1) DO BEGIN
        WRITE(' ');
        WRITE((I-1)*10:3,' !');
        FOR J := 0 TO 9 DO BEGIN
           K := (I-1)*10+J;
           IF (K>0) AND (K<=SITES)
           THEN IF STEPS[K] = NECSTEPS[K]
                THEN WRITE('Y')
                ELSE WRITE('N')
           ELSE WRITE(' ');
           END;
        WRITELN;
        END;
     WRITELN;
     WRITELN(' HAS ',LIKE:4,' COMPATIBLE SITES');
     WRITELN;
  END; (* PRINTREE *)
PROCEDURE EVALUATE (R: POINTER);
  (* DETERMINES THE "LIKELIHOOD" OF A TREE *)
  PROCEDURE CONSIDER (P: POINTER);
     (* CONSTRUCTS BITSETS AT THE NODE
         AND COUNTS HOW MANY BASE CHANGES ARE NECESSARY
         AND COUNTS HOW MANY CHARACTERS HAVE MIN NO OF STEPS *)
     VAR I : INTEGER;
         NS, RS, LS : BASET; (* TEMPORARY BASE SETS *)
     BEGIN (* CONSIDER *)
         IF NOT P^.TIP THEN
           FOR I:= 1 TO SITES DO BEGIN
              LS := P^.LEFTDESC^.STATESET[I];
              RS := P^.RTDESC^.STATESET[I];
              NS := LS * RS;
              IF NS = []
              THEN BEGIN
                 NS := LS + RS;
                 IF NS <> []
                 THEN BEGIN
                    STEPS[I] := STEPS[I] + 1;
                    END;
                 END;
              P^.STATESET[I] := NS;
              END;
     END; (* CONSIDER *)
  PROCEDURE POSTORDER ( P : POINTER );
     (* TRAVERSES A BINARY TREE, CONSIDERING A NODE'S DESCENDANTS
        BEFORE CONSIDERING THE NODE *)
     BEGIN (* POSTORDER *)
     IF P <> NIL
     THEN BEGIN
        POSTORDER (P^.LEFTDESC);
        POSTORDER (P^.RTDESC);
        CONSIDER (P);
        END
     END; (* POSTORDER *)
  BEGIN (* EVALUATE *)
     FOR I := 1 TO SITES DO STEPS[I] := 0;
     POSTORDER (R);
     LIKE := 0;
     FOR I := 1 TO SITES DO
        IF STEPS[I] = NECSTEPS [I]
        THEN LIKE := LIKE + 1;
  END; (* EVALUATE *)
PROCEDURE REARRANGE (VAR R: POINTER);
  (* TRAVERSES THE TREE (PREORDER), FINDING ANY LOCAL
     REARRANGEMENT WHICH INCREASES THE "LIKELIHOOD".
     IF TRAVERSAL SUCCEEDS IN INCREASING THE TREE'S "LIKELIHOOD"
     PROCEDURE REARRANGE RUNS TRAVERSAL AGAIN *)
  VAR SUCCESS : BOOLEAN;
  PROCEDURE TRYREARR (P : POINTER);
     (* EVALUATES ONE REARRANGEMENT OF THE TREE.
        IF THE NEW TREE HAS GREATER "LIKELIHOOD" THAN THE OLD ONE
        SETS SUCCESS = TRUE AND KEEPS THE NEW TREE.
        OTHERWISE, RESTORES THE OLD TREE *)
     VAR FROMBELOW, WHERETO, FORKNODE : POINTER;
        OLDLIKE : INTEGER;
     BEGIN (* TRYREARR *)
        IF P^.ANCESTOR <> NIL THEN BEGIN
           IF P^.ANCESTOR^.ANCESTOR <> NIL THEN BEGIN
              OLDLIKE := LIKE;
              IF P^.ANCESTOR^.LEFTDESC = P
                 THEN FROMBELOW := P^.ANCESTOR^.RTDESC
                 ELSE FROMBELOW := P^.ANCESTOR^.LEFTDESC;
              WHERETO := P^.ANCESTOR^.ANCESTOR;
              FORKNODE := P^.ANCESTOR;
              REMOVE (P, FORKNODE);
              ADD (WHERETO, P, FORKNODE);
              IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
              EVALUATE (R);
              IF LIKE <= OLDLIKE
              THEN BEGIN
                 REMOVE (P, FORKNODE);
                 ADD (FROMBELOW, P, FORKNODE);
                 IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
                 LIKE := OLDLIKE;
                 END
              ELSE BEGIN
                 SUCCESS := TRUE;
                 END;
              END;
           END;
     END; (* TRYREARR *)
  PROCEDURE PREORDER ( P : POINTER );
    (* TRAVERSES A BINARY TREE, CALLING PROCEDURE TRYREARR
       AT A NODE BEFORE CALLING TRYREARR AT ITS DESCENDANTS *)
     BEGIN (* PREORDER *)
     IF P <> NIL
     THEN BEGIN
        TRYREARR ( P );
        PREORDER (P^.LEFTDESC);
        PREORDER (P^.RTDESC);
        END;
     END; (* PREORDER *)
  BEGIN (* REARRANGE *)
     SUCCESS := TRUE;
     WHILE SUCCESS DO BEGIN
        SUCCESS := FALSE;
        PREORDER (R);
        END;
  END; (* REARRANGE *)
PROCEDURE MINCOMP (SPRIME : INTEGER);
  (* FINDS FOR EACH SITE THE MINIMUM NUMBER OF STEPS NECESSARY
     FOR FIRST SPRIME SPECIES AND PUTS THESE NUMBERS IN NECSTEPS *)
  VAR I, J, K, L, N : INTEGER;
       B : BASES;
       S : BASET;
       ALLOWABLE, DELETED : BOOLEAN;
  BEGIN (* MINCOMP *)
     FOR I := 1 TO SITES DO NECSTEPS[I] := 3;
     FOR N := 0 TO 15 DO BEGIN
        S := [];
        L := -1;
        K := N;
        FOR B := A TO U DO BEGIN
           IF (K MOD 2) = 1
           THEN BEGIN
              S := S + [B];
              L := L + 1;
              END;
           K := K DIV 2;
           END;
        FOR J := 1 TO SITES DO BEGIN
           ALLOWABLE := TRUE;
           I := 1;
           WHILE ALLOWABLE AND (I<=SPRIME) DO BEGIN
              IF TREENODE[I]^.STATESET[J] <> []
              THEN IF (TREENODE[I]^.STATESET[J] * S) = []
                   THEN ALLOWABLE := FALSE;
              I := I + 1;
              END;
           IF ALLOWABLE
           THEN IF L < NECSTEPS[J]
                THEN NECSTEPS[J] := L;
           END;
        END;
        FOR J := 1 TO SITES DO BEGIN
           DELETED := FALSE;
           FOR I := 1 TO SPRIME DO
              IF TREENODE[I]^.STATESET[J] = []
              THEN DELETED := TRUE;
           IF DELETED
           THEN NECSTEPS[J] := NECSTEPS[J] + 1;
           END;
   END; (* MINCOMP *)
PROCEDURE ADDER (VAR R: POINTER; VAR NODEARRAY : POINTARRAY; SPP :
                 INTEGER; VAR LIKE: INTEGER);
  (* CONSTRUCTS A BINARY TREE FROM THE POINTERS IN NODEARRAY.
     ADDS EACH NODE AT LOCATION WHICH YIELDS HIGHEST "LIKELIHOOD"
     THEN REARRANGES THE TREE FOR GREATEST "LIKELIHOOD" *)
  VAR I : INTEGER;
     THERE, ITEM : POINTER;
     BESTYET : INTEGER;
  PROCEDURE TRYADD (P: POINTER);
     (* TEMPORARILY ADDS ONE FORK AND ONE TIP TO THE TREE.
        IF THE LOCATION WHERE THEY ARE ADDED YIELDS GREATER
        "LIKELIHOOD" THAN OTHER LOCATIONS TESTED UP TO THAT TIME,
        THEN KEEPS THAT LOCATION AS BESTYET *)
     VAR NUFORK : POINTER;
     BEGIN (* TRYADD *)
        NUFORK := NODEARRAY[SPP+I-1];
        ADD (P, ITEM, NUFORK);
        IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
        EVALUATE (R);
        REMOVE (ITEM, NUFORK);
        IF NUFORK = R THEN R := P;
        IF LIKE > BESTYET THEN BEGIN
           BESTYET := LIKE;
           THERE := P;
           END;
     END; (* TRYADD *)
  PROCEDURE PREORDER ( P : POINTER );
     (* TRAVERSES A BINARY TREE, CALLING PROCEDURE TRYADD AT A
        NODE BEFORE CALLING TRYADD AT ITS DESCENDANTS *)
     BEGIN (* PREORDER *)
     IF P <> NIL
     THEN BEGIN
        TRYADD ( P );
        PREORDER (P^.LEFTDESC);
        PREORDER (P^.RTDESC);
        END
     END; (* PREORDER *)
  BEGIN (* ADDER *)
     ADD (NODEARRAY[2], NODEARRAY[1], NODEARRAY[SPP+1]);
     R := NODEARRAY[SPP+1];
     FOR I:= 3 TO SPP DO BEGIN
        MINCOMP (I);
        ITEM := NODEARRAY [I];
        BESTYET := 32000;
        THERE := R;
        PREORDER (R);
        ADD (THERE, NODEARRAY[I], NODEARRAY[SPP+I-1]);
        IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
        EVALUATE(R);
        REARRANGE (R);
        EVALUATE(R);
        PRINTREE(R);
        END;
  END; (* ADDER *)
BEGIN (* BASE CHANGE COMPATIBILITY BY UPHILL SEARCH *)
  RESET(INFILE);
  (* READS IN SPP, SEQUENCES, AND THE DATA; THEN CALLS ADDER TO
     CONSTRUCT THE TREE AND PRINTREE TO PRINT IT OUT *)
  WRITELN;
  WRITELN(' ','DNA BASE CHANGE COMPATIBILITY ALGORITHM');
  WRITELN;
  READLN( INFILE, SPP, SITES );
  WRITELN(' ',SPP:2,' SPECIES, ',SITES:3,'   BASE SITES');
  WRITELN;
  NONODES := 2 * SPP - 1;
  FOR I:=1 TO NONODES DO BEGIN
     NEW (TREENODE[I]);
     TREENODE[I]^.LEFTDESC := NIL;
     TREENODE[I]^.RTDESC := NIL;
     TREENODE[I]^.ANCESTOR := NIL;
     TREENODE[I]^.INDEX := I;
     IF I <= SPP
     THEN BEGIN
        TREENODE[I]^.TIP := TRUE;
        FOR J := 1 TO NMLNGTH DO READ (INFILE,NAME[I,J]);
        WRITE(' ');
        FOR J := 1 TO NMLNGTH DO WRITE(NAME[I,J]);
        WRITE(' ');
        FOR J := 1 TO SITES DO BEGIN
           IF EOLN(INFILE) THEN READLN(INFILE);
           READ (INFILE,CHARSTATE);
           IF CHARSTATE > 'Z'
           THEN CHARSTATE := CHR(ORD(CHARSTATE) - 32); (* ASSUMES ASCII CODE *)
           WRITE(CHARSTATE);
           IF J MOD 3 = 0 THEN WRITE(' ');
           IF J MOD 51 = 0
           THEN BEGIN
              WRITELN;
              WRITE(' ');
              FOR K := 1 TO NMLNGTH DO WRITE(' ');
              END;
           CASE CHARSTATE OF
              'A' : NS := [A];
              'C' : NS := [C];
              'G' : NS := [G];
              'U' : NS := [U];
              'T' : NS := [U];
              'R' : NS := [A, G];
              'Y' : NS := [C, U];
              'H' : NS := [A, C, U];
              'N', 'X', '?' : NS := [A, C, G, U];
              'O' : NS := [];
              ' ' : NS := []
              END;
        TREENODE[I]^.STATESET[J] := NS;
        END;
        READLN(INFILE);
        WRITELN;
        END
     ELSE TREENODE[I]^.TIP := FALSE;
     END;
  LIKE := 0;
  ADDER (ROOT, TREENODE, SPP, LIKE);
  (*  CLOSE(INFILE);  -- MAY BE NEEDED HERE *)
END.
