PROGRAM POLYM(INFILE, OUTPUT);
  (* VERSION 1.5 *)
CONST MAXSZ = 100;  (* SIZE OF POINTER ARRAY.  >= 2*MAXSP - 1  *)
  MAXSP = 50;       (* MAXIMUM NUMBER OF SPECIES. CAN'T BE >100 *)
  BITS = 30;        (* MAX. NUMBER OF ELEMENTS IN A PASCAL SET *)
  WRDS = 7;         (* WRDS * BITS >= MAXCHR                   *)
  NMLNGTH = 10;     (* NUMBER OF CHARACTERS IN SPECIES NAME    *)
  MAXCHR = 200;     (* MAX. NUMBER CHARACTERS PER SPECIES *)
TYPE POINTER = ^NODE;           (* NODES WILL FORM BINARY TREE *)
  POINTARRAY = ARRAY [1..MAXSZ] OF POINTER;
  BITRANGE = 1..BITS;
  BITSET = SET OF BITRANGE;
  NODE = RECORD      (* DESCRIBES A TIP SPECIES OR AN ANCESTOR *)
     ANCESTOR, LEFTDESC, RTDESC : POINTER;
     INDEX : INTEGER;
     TIP : BOOLEAN;        (* PRESENT SPECIES ARE TIPS OF TREE *)
     STATEMAX : ARRAY [1..WRDS] OF BITSET;
     STATEMIN : ARRAY [1..WRDS] OF BITSET;
     END;
VAR ROOT : POINTER;
  INFILE : TEXT;
  K : BITRANGE;
  NUMSTEPS, BESTEPS : ARRAY[1..MAXCHR] OF INTEGER;
  I, J, L, SPP, WORDS, CHARS, NONODES : INTEGER;
    (* SPP = NUMBER OF SPECIES; WORDS = NUMBER OF WORDS NEEDED
       TO REPRESENT CHARACTERS OF ONE ORGANISM;
       CHARS = NUMBER OF CHARACTERS;
       NONODES = NUMBER OF NODES IN TREE *)
  LIKE : INTEGER;                    (* "LIKELIHOOD" OF A TREE *)
  TREENODE : POINTARRAY;      (* POINTERS TO ALL NODES IN TREE *)
  NAME : ARRAY[1..MAXSP, 1..NMLNGTH] OF CHAR; (*NAMES OF SPECIES*)
  CHARSTATE : CHAR;        (* POSSIBLE STATES ARE '0', '1', 'P', 'B', and '?'*)
PROCEDURE ADD (VAR BELOW, NEWTIP, NEWFORK : POINTER);
  (* INSERTS THE NODES NEWFORK AND ITS LEFT DESCENDANT, NEWTIP,
     TO THE TREE.  BELOW BECOMES NEWFORK'S RIGHT DESCENDANT *)
  BEGIN (* ADD *)
     NEWFORK^.ANCESTOR := BELOW^.ANCESTOR;
     BELOW^.ANCESTOR := NEWFORK;
     NEWTIP^.ANCESTOR := NEWFORK;
     NEWFORK^.RTDESC := BELOW;
     NEWFORK^.LEFTDESC := NEWTIP;
     IF NEWFORK^.ANCESTOR <> NIL THEN BEGIN
        IF NEWFORK^.ANCESTOR^.LEFTDESC = BELOW
        THEN NEWFORK^.ANCESTOR^.LEFTDESC := NEWFORK
        ELSE NEWFORK^.ANCESTOR^.RTDESC := NEWFORK;
        END
  END; (* ADD *)
PROCEDURE REMOVE (VAR ITEM, FORK : POINTER);
  (* REMOVES NODES ITEM AND ITS ANCESTOR, FORK, FROM THE TREE.
     THE NEW DESCENDANT OF FORK'S ANCESTOR IS MADE TO BE
     FORK'S SECOND DESCENDANT (OTHER THAN ITEM).
     ALSO RETURNS POINTERS TO THE DELETED NODES, ITEM AND FORK *)
  VAR OTHER, BELOW : POINTER;
  BEGIN (* REMOVE *)
     IF ITEM^.ANCESTOR = NIL
     THEN FORK := NIL
     ELSE BEGIN
        FORK := ITEM^.ANCESTOR;
        IF ITEM = FORK^.LEFTDESC
        THEN OTHER := FORK^.RTDESC
        ELSE OTHER := FORK^.LEFTDESC;
        BELOW := FORK^.ANCESTOR;
        OTHER^.ANCESTOR := BELOW;
        IF BELOW <> NIL THEN BEGIN
           IF BELOW^.LEFTDESC = FORK
           THEN BELOW^.LEFTDESC := OTHER
           ELSE BELOW^.RTDESC := OTHER;
           FORK^.ANCESTOR := NIL;
           END;
        FORK^.LEFTDESC := NIL;
        FORK^.RTDESC := NIL;
        ITEM^.ANCESTOR := NIL;
        END;
  END; (* REMOVE *)
PROCEDURE PRINTREE (P: POINTER);
  (* PRINTS A BINARY ROOTED TREE *)
  CONST C1=2; C2=3;
  VAR HORIZ: ARRAY[1..MAXSZ] OF INTEGER;
     ALPHA: ARRAY[1..100] OF CHAR;
     I, J, PAST, INDX, OLDIND, PLACE1, PLACE2, C22 : INTEGER;
  PROCEDURE MIXORDER (R: POINTER);
     (* THIS IS A MIXED-ORDER TREE TRAVERSAL WHICH DOES MOST OF
        THE WORK OF PRINTREE *)
     VAR I,J : INTEGER;
     BEGIN (* MIXORDER *)
        IF NOT R^.TIP THEN
           MIXORDER (R^.LEFTDESC);
        INDX := R^.INDEX;
        IF R^.TIP THEN BEGIN
           IF PAST > 0 THEN BEGIN
              WRITE('   ');
              FOR I:=1 TO C22*(PAST-1)+1 DO WRITE(ALPHA[I]);
              FOR I:=1 TO NMLNGTH DO WRITE(NAME[OLDIND,I]);
              WRITELN;
              FOR I:= C22*PAST DOWNTO 2 DO BEGIN
                 IF ALPHA[I] = '-' THEN ALPHA[I] := ' ';
                 IF ALPHA[I] = '*' THEN BEGIN
                    IF ALPHA[I-1] = ' '
                    THEN ALPHA[I] := '!'
                    ELSE ALPHA[I] := ' ';
                    END;
                 END;
              IF ALPHA[1] = '*' THEN ALPHA[1] := '!';
              FOR I:=1 TO C1-1 DO BEGIN
                 WRITE('   ');
                 FOR J:=1 TO C22*PAST DO WRITE(ALPHA[J]);
                 WRITELN;
                 END
              END;
           OLDIND := INDX;
           ALPHA[C22*PAST + 1] := '!';
           FOR I:= C22*PAST+2 TO C22*(PAST+1) DO ALPHA[I] := ' ';
           PAST := PAST +1;
           PLACE2 := PAST;
           HORIZ[INDX] := PAST;
           END;
        IF NOT R^.TIP THEN BEGIN
           HORIZ[INDX] := HORIZ[R^.LEFTDESC^.INDEX];
           MIXORDER (R^.RTDESC);
           END;
        IF R^.ANCESTOR <> NIL THEN BEGIN
           IF R = R^.ANCESTOR^.RTDESC THEN BEGIN
              PLACE1 := HORIZ[R^.ANCESTOR^.INDEX];
              FOR I:= C22*(PLACE1-1)+2 TO C22*(PLACE2-1)+1 DO
                 IF ALPHA[I] <> '*' THEN ALPHA[I] := '-';
              ALPHA[C22*(PLACE1-1)+1] := '*';
              PLACE2 := PLACE1;
              END;
        END;
     END; (* MIXORDER *)
  BEGIN (* PRINTREE *)
     C22 := 100 DIV SPP;
     IF C22 > C2 THEN C22 := C2;
     PAST := 0;
     MIXORDER (P);
     WRITE('   ');
     ALPHA[1] := '*';
     FOR I:=1 TO C22*(PAST-1)+1 DO WRITE(ALPHA[I]);
     FOR I:= 1 TO NMLNGTH DO WRITE(NAME[INDX,I]);
     WRITELN;
     WRITELN('  /');
     WRITELN(' /');
     WRITELN;
     WRITELN(' REQUIRES A TOTAL OF ',-LIKE:7,'  RETENTIONS OF POLYMORPHISM');
     WRITELN;
     WRITELN(' POLYMORPHIC SEGMENTS IN EACH CHARACTER:');
     WRITE('       ');
     FOR I := 0 TO 9 DO WRITE(I:3);
     WRITELN;
     WRITELN('      *------------------------------');
     FOR I := 0 TO (CHARS DIV 10) DO BEGIN
        WRITE(10*I:6);
        WRITE('!');
        FOR J := 0 TO 9 DO BEGIN
           IF 10*I+J = 0
           THEN WRITE('   ')
           ELSE IF 10*I+J <= CHARS
              THEN WRITE(BESTEPS[10*I+J]:3)
              ELSE WRITE('   ');
           END;
        WRITELN;
        END;
        WRITELN;
        WRITELN;
   END; (* PRINTREE *)
  PROCEDURE EVALUATE (R: POINTER);
  (* DETERMINES THE NUMBER OF STEPS NEEDED FOR A TREE. THIS IS
     THE MINIMUM NUMBER OF STEPS NEEDED TO EVOLVE CHARS ON THIS TREE *)
  VAR I : INTEGER;
    SUM : INTEGER;
  PROCEDURE FILLIN (P: POINTER);
   (* SETS UP FOR EACH NODE IN THE TREE TWO STATESETS. ONE, STATEMAX,
     IS THE SET OF CHARACTERS WHICH MIGHT BE IN STATE 1 AT THIS POINT IN
     THE TREE IN A "MOST PARSIMONIOUS" RECONSTRUCTION OF STATES AT
     INTERMEDIATE NODES. THE OTHER, STATEMIN, SHOWS WHICH CHARACTERS
     MIGHT BE IN STATE 0 *)
     VAR I : 1..WRDS;
     BEGIN (* FILLIN *)
        FOR I := 1 TO WORDS DO BEGIN
           P^.STATEMAX[I] := P^.LEFTDESC^.STATEMAX[I]
                              + P^.RTDESC^.STATEMAX[I];
           P^.STATEMIN[I] := P^.LEFTDESC^.STATEMIN[I]
                              + P^.RTDESC^.STATEMIN[I];
           END;
     END; (* FILLIN *)
  PROCEDURE COUNT (P : POINTER);
     (* COUNTS THE NUMBER OF STEPS IN A BRANCH OF THE TREE *)
     VAR I, J, L : INTEGER;
        STEPS : ARRAY[1..WRDS] OF BITSET;
     BEGIN (* COUNT *)
        FOR I := 1 TO WORDS DO
           STEPS[I] := P^.STATEMAX[I] * P^.STATEMIN[I]
                * P^.ANCESTOR^.STATEMAX[I] * P^.ANCESTOR^.STATEMIN[I];
        J := 1;
        L := 0;
        FOR I := 1 TO CHARS DO BEGIN
           L := L + 1;
           IF L>BITS THEN BEGIN
              L := 1;
              J := J + 1;
              END;
           IF L IN STEPS[J]
           THEN NUMSTEPS[I] := NUMSTEPS[I] + 1;
           END;
     END; (* COUNT *)
  PROCEDURE POSTORDER ( P : POINTER );
     (* TRAVERSES A BINARY TREE, CALLING PROCEDURE FILLIN AT
        A NODE'S DESCENDANTS BEFORE CALLING FILLIN AT THE NODE *)
     BEGIN (* POSTORDER *)
        IF NOT P^.TIP
        THEN BEGIN
           POSTORDER(P^.LEFTDESC);
           POSTORDER(P^.RTDESC);
           FILLIN(P);
           END;
     END; (* POSTORDER *)
  PROCEDURE SNIP (P : POINTER);
     (* WORK OUT NUMBER OF STATES ACTUALLY PRESENT AND COUNT POLYMORPHISMS *)
     VAR I : 1..WRDS;
     BEGIN (* SNIP *)
        IF NOT P^.TIP
        THEN BEGIN
           FOR I := 1 TO WORDS DO
              P^.STATEMAX[I] := P^.LEFTDESC^.STATEMAX[I]
                                  * P^.RTDESC^.STATEMAX[I];
           IF P^.ANCESTOR <> NIL
           THEN BEGIN
              FOR I := 1 TO WORDS DO
                 P^.STATEMAX[I] := P^.STATEMAX[I] + (P^.LEFTDESC^.STATEMAX[I]
                          + P^.RTDESC^.STATEMAX[I]) * P^.ANCESTOR^.STATEMAX[I];
              END;
        END;
        IF P^.ANCESTOR <> NIL THEN COUNT (P);
     END; (* SNIP *)
  PROCEDURE PREORDER ( P : POINTER);
      (* TRAVERSES A BINARY TREE, CALLING PROCEDURE SNIP AT A NODE'S
         ANCESTORS BEFORE CALLING IT AT THE NODE *)
      BEGIN (* PREORDER *)
        SNIP(P);
        IF NOT P^.TIP
        THEN BEGIN
           PREORDER(P^.LEFTDESC);
           PREORDER(P^.RTDESC);
           END;
      END; (* PREORDER *)
  BEGIN (* EVALUATE *)
     SUM := 0;
     FOR I := 1 TO CHARS DO NUMSTEPS[I] := 0;
     POSTORDER (R);
     PREORDER (R);
     FOR I := 1 TO CHARS DO
        SUM := SUM + NUMSTEPS[I];
     LIKE := -SUM;
  END; (* EVALUATE *)
PROCEDURE REARRANGE (VAR R: POINTER; VAR LIKE : INTEGER);
  (* TRAVERSES THE TREE (PREORDER), FINDING ANY LOCAL
     REARRANGEMENT WHICH INCREASES THE "LIKELIHOOD".
     IF TRAVERSAL SUCCEEDS IN INCREASING THE TREE'S "LIKELIHOOD",
     PROCEDURE REARRANGE RUNS TRAVERSAL AGAIN *)
  VAR SUCCESS : BOOLEAN;
  PROCEDURE TRYREARR (P : POINTER);
     (* EVALUATES ONE REARRANGEMENT OF THE TREE.
        IF THE NEW TREE HAS GREATER "LIKELIHOOD" THAN THE OLD ONE
        SETS SUCCESS = TRUE AND KEEPS THE NEW TREE.
        OTHERWISE, RESTORES THE OLD TREE *)
     VAR FROMBELOW, WHERETO, FORKNODE : POINTER;
        OLDLIKE : INTEGER;
     BEGIN (* TRYREARR *)
        IF P^.ANCESTOR <> NIL THEN BEGIN
           IF P^.ANCESTOR^.ANCESTOR <> NIL THEN BEGIN
              OLDLIKE := LIKE;
              IF P^.ANCESTOR^.LEFTDESC = P
                 THEN FROMBELOW := P^.ANCESTOR^.RTDESC
                 ELSE FROMBELOW := P^.ANCESTOR^.LEFTDESC;
              WHERETO := P^.ANCESTOR^.ANCESTOR;
              FORKNODE := P^.ANCESTOR;
              REMOVE (P, FORKNODE);
              ADD (WHERETO, P, FORKNODE);
              IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
              EVALUATE (R);
              IF LIKE <= OLDLIKE
              THEN BEGIN
                 REMOVE (P, FORKNODE);
                 ADD (FROMBELOW, P, FORKNODE);
                 IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
                 LIKE := OLDLIKE;
                 END
              ELSE BEGIN
                 SUCCESS := TRUE;
                 BESTEPS := NUMSTEPS;
                 END;
              END;
           END;
     END; (* TRYREARR *)
  PROCEDURE PREORDER ( P : POINTER );
     (* TRAVERSES A BINARY TREE, CALLING PROCEDURE TRYREARR
        AT A NODE BEFORE CALLING TRYREARR AT ITS DESCENDANTS *)
     BEGIN (* PREORDER *)
     IF P <> NIL
     THEN BEGIN
        TRYREARR ( P );
        PREORDER (P^.LEFTDESC);
        PREORDER (P^.RTDESC);
        END;
     END; (* PREORDER *)
  BEGIN (* REARRANGE *)
     SUCCESS := TRUE;
     WHILE SUCCESS DO BEGIN
        SUCCESS := FALSE;
        PREORDER (R);
        END;
  END; (* REARRANGE *)
PROCEDURE ADDER (VAR R: POINTER; VAR NODEARRAY : POINTARRAY; SPP:
                 INTEGER; VAR LIKE: INTEGER);
  (* CONSTRUCTS A BINARY TREE FROM THE POINTERS IN NODEARRAY.
     ADDS EACH NODE AT LOCATION WHICH YIELDS HIGHEST "LIKELIHOOD"
     THEN REARRANGES THE TREE FOR GREATEST "LIKELIHOOD" *)
  VAR I : INTEGER;
     THERE, ITEM : POINTER;
     BESTYET : INTEGER;
  PROCEDURE TRYADD (P: POINTER);
     (* TEMPORARILY ADDS ONE FORK AND ONE TIP TO THE TREE.
        IF THE LOCATION WHERE THEY ARE ADDED YIELDS GREATER
        "LIKELIHOOD" THAN OTHER LOCATIONS TESTED UP TO THAT TIME,
        THEN KEEPS THAT LOCATION AS THERE *)
     VAR NUFORK : POINTER;
     BEGIN (* TRYADD *)
        NUFORK := NODEARRAY[SPP+I-1];
        ADD (P, ITEM, NUFORK);
        IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
        EVALUATE (R);
        REMOVE (ITEM, NUFORK);
        IF NUFORK = R THEN R := P;
        IF LIKE > BESTYET THEN BEGIN
           BESTYET := LIKE;
           THERE := P;
           BESTEPS := NUMSTEPS;
           END;
     END; (* TRYADD *)
  PROCEDURE PREORDER ( P : POINTER );
     (* TRAVERSES A BINARY TREE, CALLING PROCEDURE TRYADD
        AT A NODE BEFORE CALLING TRYADD AT ITS DESCENDANTS *)
     BEGIN (* PREORDER *)
     IF P <> NIL
     THEN BEGIN
        TRYADD ( P );
        PREORDER (P^.LEFTDESC);
        PREORDER (P^.RTDESC);
        END;
     END; (* PREORDER *)
  BEGIN (* ADDER *)
     ADD (NODEARRAY[2], NODEARRAY[1], NODEARRAY[SPP+1]);
     R := NODEARRAY[SPP+1];
     EVALUATE (R);
     FOR I:= 3 TO SPP DO BEGIN
        ITEM := NODEARRAY [I];
        BESTYET := -32000;
        THERE := R;
        PREORDER (R);
        ADD (THERE, NODEARRAY[I], NODEARRAY[SPP+I-1]);
        IF R^.ANCESTOR <> NIL THEN R := R^.ANCESTOR;
        LIKE := BESTYET;
        REARRANGE (R, LIKE);
        PRINTREE (R);
        END;
  END; (* ADDER *)
BEGIN (* POLYMORPHISM PARSIMONY BY UPHILL SEARCH *)
  (* READS IN SPP, CHARS, AND THE DATA; THEN CALLS ADDER TO
     CONSTRUCT THE TREE AND PRINTREE TO PRINT IT OUT *)
  RESET(INFILE);
  WRITELN;
  WRITELN(' ','POLYMORPHISM PARSIMONY ALGORITHM');
  WRITELN;
  READ( INFILE, SPP, CHARS );
  WRITELN(' ',SPP:2,' SPECIES, ',CHARS:3,'  CHARACTERS');
  WRITELN;
  WORDS := (CHARS DIV BITS) + 1;
  NONODES := 2 * SPP - 1;
  READLN(INFILE);
  FOR I:=1 TO NONODES DO BEGIN
     NEW (TREENODE[I]);
     TREENODE[I]^.LEFTDESC := NIL;
     TREENODE[I]^.RTDESC := NIL;
     TREENODE[I]^.ANCESTOR := NIL;
     TREENODE[I]^.INDEX := I;
     IF I <= SPP
     THEN BEGIN
        TREENODE[I]^.TIP := TRUE;
        FOR J := 1 TO NMLNGTH DO READ (INFILE,NAME[I,J]);
        WRITE(' ');
        FOR J := 1 TO NMLNGTH DO WRITE(NAME[I,J]);
        WRITE('   ');
        FOR J := 1 TO WORDS DO BEGIN
           TREENODE[I]^.STATEMAX[J] := [];
           TREENODE[I]^.STATEMIN[J] := [];
           END;
        FOR J := 1 TO CHARS DO BEGIN
           K := ((J-1) MOD BITS) + 1;
           L := ((J-1) DIV BITS) + 1;
           IF EOLN(INFILE) THEN READLN(INFILE);
           READ (INFILE,CHARSTATE);
           IF (J <> 1) AND ((J-1) MOD 50 = 0) THEN BEGIN
              WRITELN;
              WRITE('               ');
              END;
           WRITE(CHARSTATE);
           IF (J MOD 5) = 0 THEN WRITE(' ');
           IF CHARSTATE IN ['1', 'B', 'P'] THEN
              TREENODE[I]^.STATEMAX[L]
               := TREENODE[I]^.STATEMAX[L] + [K];
           IF CHARSTATE IN ['0', 'B', 'P'] THEN
              TREENODE[I]^.STATEMIN[L]
               := TREENODE[I]^.STATEMIN[L] + [K];
           END;
        READLN(INFILE);
        WRITELN;
        END
     ELSE TREENODE[I]^.TIP := FALSE;
     END;
  WRITELN;
  WRITELN;
  LIKE := -32000;
  ADDER (ROOT, TREENODE, SPP, LIKE);
  (*  CLOSE(INFILE);  -- MAY HAVE TO BE PUT HERE *)
END. (* POLYMORPHISM PARSIMONY BY UPHILL SEARCH *)
